from __future__ import annotations
from typing import Optional

import os
import sys
import socket
import json
import pickle
import subprocess
from pathlib import Path

DEFAULT_ADDR = '127.0.0.1'
DEFAULT_PORT_LINEPLOTTER = 8765
DEFAULT_PORT_IMAGEPLOTTER = 8766


def plot_lines(*args,
               xlabel: Optional[str] = None, ylabel: Optional[str] = None,
               windowsize: Optional[tuple[int, int]] = None,
               addr: str = DEFAULT_ADDR,
               port: int = DEFAULT_PORT_LINEPLOTTER):
    """
    args:
        - ydata
        - [ydata]
        - xdata, ydata
        - xdata, [ydata]
        - xdata, ydata1, ydata2
        - xdata, ydata1, ydata2, ...
    """
    _ping_or_launch_lineplotter(addr, port)

    _send_data(args, addr, port)

    attrs = dict(xlabel=xlabel, ylabel=ylabel, windowsize=windowsize)
    _send_attrs(addr, port, attrs)


def plot_image(img,
               xlabel: Optional[str] = None, ylabel: Optional[str] = None,
               windowsize: Optional[tuple[int, int]] = None,
               addr: str = DEFAULT_ADDR,
               port: int = DEFAULT_PORT_IMAGEPLOTTER):
    """Plot an image

    attrs:
    - img, 2d array_like

    TODO: xaxis, yaxisを受け入れる
    """
    _ping_or_launch_imagplotter(addr, port)

    _send_data(img, addr, port)

    attrs = dict(xlabel=xlabel, ylabel=ylabel, windowsize=windowsize)
    _send_attrs(addr, port, attrs)


def plot_image_and_lines(img,
                         xlabel: Optional[str] = None, ylabel: Optional[str] = None,
                         windowsize: Optional[tuple[int, int]] = None,
                         addr: str = DEFAULT_ADDR,
                         port_image: int = DEFAULT_PORT_IMAGEPLOTTER,
                         port_lines: int = DEFAULT_PORT_LINEPLOTTER):
    """Plot an image, and plot each row of the image
    """
    plot_image(img, xlabel=xlabel, ylabel=ylabel, windowsize=windowsize,
               addr=addr, port=port_image)
    plot_lines(img, xlabel=xlabel,
               addr=addr, port=port_lines)


def _ping_or_launch_lineplotter(addr: str, port: int):
    """lineplotterがあるか確認、起動してなければ起動する
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((addr, port))
            header = json.dumps({'type': 'ping'}).encode('utf-8')
            s.send(header)
    except ConnectionRefusedError:
        fn_entry = Path(__file__).parent / 'entry_points/lineplotter.py'
        _ = subprocess.Popen([_get_executable(), str(fn_entry.absolute()),
                              '--addr', addr, '--port', str(port)],
                             stdout=subprocess.DEVNULL,
                             stderr=subprocess.DEVNULL)


def _ping_or_launch_imagplotter(addr: str, port: int):
    """imageplotterがあるか確認、起動してなければ起動する
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((addr, port))
            header = json.dumps({'type': 'ping'}).encode('utf-8')
            s.send(header)
    except ConnectionRefusedError:
        fn_entry = Path(__file__).parent / 'entry_points/imageplotter.py'
        _ = subprocess.Popen([_get_executable(), str(fn_entry.absolute()),
                              '--addr', addr, '--port', str(port)],
                             stdout=subprocess.DEVNULL,
                             stderr=subprocess.DEVNULL)


def _send_data(v, addr: str, port: int):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((addr, port))

        data = pickle.dumps(v)
        header = json.dumps({'size': len(data), 'type': 'data'}).encode('utf-8')

        s.send(header)
        _ = s.recv(2048)
        s.sendall(data)


def _send_attrs(addr: str, port: int, attrs: dict):
    if not any(attrs.values()):
        return

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((addr, port))

        data = pickle.dumps(attrs)
        header = json.dumps({'size': len(data), 'type': 'attr'}).encode('utf-8')

        s.send(header)
        _ = s.recv(2048)
        s.sendall(data)


def _get_executable() -> str:
    KEY_EXE = 'SOCKETPLOTTER_PYTHON_EXECUTABLE'
    if KEY_EXE in os.environ:
        p = Path(os.environ[KEY_EXE])
        if p.exists():
            return str(p)

    return sys.executable



from socket_plotter import plot_lines, plot_image, plot_image_and_lines
import numpy as np

# for line plot
xdata = np.arange(100)
ydata = np.random.randn(100)
plot_lines(xdata, ydata)

# for image plot
img = np.random.randn(100, 100)
plot_image(img)

# for both, the image and each row of the image are displayed.
img = np.random.randn(100, 100)
plot_image_and_lines(img)



from __future__ import annotations

from PySide2.QtWidgets import QApplication
import numpy as np
import pyqtgraph as pg

from .receiver import QThreadReceiver


class LinePlotter():
    DEFAULT_SIZE = (600, 400)
    DEFAULT_WINDOW_TITLE = 'Line Plotter'

    def __init__(self, addr: str, port: int):
        self.app = QApplication([])
        self.win = pg.GraphicsLayoutWidget(title=self.DEFAULT_WINDOW_TITLE)
        self.win.resize(*self.DEFAULT_SIZE)
        self.win.show()

        self.plotitem: pg.PlotItem = self.win.addPlot()
        self.plotitem.showGrid(x=True, y=True)
        self.plots: list[pg.PlotDataItem] = []

        # TODO: QThread の適切な終了方法がわからん
        #       `QThread: Destroyed while thread is still running` と怒られてしまう
        # self.app.aboutToQuit.conncect(self.receiver.stop()) とかやると別の怒られが起きる
        # とりあえず放置。。。
        self.receiver = QThreadReceiver(addr, port)
        self.receiver.sigData.connect(self.draw_unpack)
        self.receiver.sigAttr.connect(self.set_attributes)
        self.receiver.sigError.connect(self.clear)
        self.receiver.start()

    def set_attributes(self, attrs: dict):
        if attrs.get('xlabel', None):
            self.plotitem.setLabel('bottom', attrs['xlabel'])
        if attrs.get('ylabel', None):
            self.plotitem.setLabel('left', attrs['ylabel'])
        if attrs.get('windowsize', None):
            self.win.resize(*attrs['windowsize'])

    def draw_unpack(self, args):
        self.draw(*args)

    def clear(self):
        for p in self.plots:
            p.setData([], [])

    def draw(self, *args):
        """
        args:
            - ydata
            - [ydata]
            - xdata, ydata
            - xdata, [ydata]
            - xdata, ydata1, ydata2
            - xdata, ydata1, ydata2, ...
        """
        xdata = ydata = None
        if len(args) == 1:
            ydata = args[0]
        elif len(args) == 2:
            xdata, ydata = args
        else:  # len(args) > 2
            xdata = args[0]
            ydata = args[1:]

        vec = np.array(ydata)
        if len(vec.shape) == 1:
            vec = [vec]

        if xdata is None:
            xdata = np.arange(len(vec[0]))

        if len(self.plots) < len(vec):
            for _ in range(len(vec) - len(self.plots)):
                p = self.plotitem.plot()
                self.plots.append(p)

        self.clear()
        for i, (p, v) in enumerate(zip(self.plots, vec)):
            p.setPen(pg.mkPen(i, hues=max(len(vec), 9)))
            p.setData(xdata, v)



from __future__ import annotations
from typing import Any

import socket
import json
import pickle

from PySide2 import QtCore


class QThreadReceiver(QtCore.QThread):
    """
    1. 接続を受けたら、以下のようなdict (header) をjsonで受け取る
        {'size': int, 'type': Literal['data', 'data_json', 'attr', 'ping']}
        - type==ping なら何もせずに次の接続をまつ
    2. `A header was received.` と返す
    3. (1)で受けたサイズだけデータを受け取る
    4. type==(data|data_json) ならsigDataでui側へ渡す。type==attrならsigAttrでui側へ渡す
      - data|attr だったらデータをpickleで受け取る
      - data_json だった場合はデータをjsonで受け取る

    途中でエラーしたら sigErrorを発してui側へ通知
    """
    buffer_size = 2048
    timeout = 0.1

    sigData = QtCore.Signal(object)
    sigAttr = QtCore.Signal(object)
    sigError = QtCore.Signal()

    def __init__(self, addr: str, port: int, parent=None) -> None:
        super().__init__(parent)

        self._mutex = QtCore.QMutex()
        self._flg_listen = True  # これがTrueの間は受け付け続ける

        self.addr_port = (addr, port)

    def stop(self) -> None:
        with QtCore.QMutexLocker(self._mutex):
            self._flg_listen = False

    def run(self) -> None:
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.bind(self.addr_port)
        self.s.listen(1)
        self.s.settimeout(self.timeout)

        while self._flg_listen:
            try:
                type, dat = self._recv()
                if type in ('data', 'data_json'):
                    self.sigData.emit(dat)
                elif type == 'attr':
                    self.sigAttr.emit(dat)
                elif type == 'ping':
                    pass
                else:
                    self.sigError.emit()

            except socket.timeout:
                continue
            except ConnectionError:
                self.sigError.emit()
            except pickle.UnpicklingError:
                self.sigError.emit()
            except json.JSONDecodeError:
                self.sigError.emit()

        self.s.close()

    def _recv(self) -> tuple[str, Any]:
        conn, _ = self.s.accept()
        with conn:
            header_bytes = conn.recv(self.buffer_size)
            header = json.loads(header_bytes)
            if header['type'] == 'ping':
                return 'ping', None

            # receiving body
            conn.send(b'A header was received.')

            databuf = bytearray(header['size'])
            conn.recv_into(databuf)

        if header['type'] == 'data_json':
            return header['type'], json.loads(databuf)
        else:
            return header['type'], pickle.loads(databuf)



from __future__ import annotations

from PySide2.QtWidgets import QApplication
import numpy as np
import pyqtgraph as pg

from .receiver import QThreadReceiver


class ImagePlotter():
    DEFAULT_SIZE = (600, 400)
    DEFAULT_TITLE = 'Image Plotter'

    def __init__(self, addr: str, port: int):
        self.app = QApplication([])
        self.win = pg.GraphicsLayoutWidget(title=self.DEFAULT_TITLE)
        self.win.resize(*self.DEFAULT_SIZE)
        self.win.show()

        self.imageitem = pg.ImageItem()
        self.imageitem.setOpts(axisOrder='row-major')

        self.viewbox = self.win.addViewBox()
        self.viewbox.setAspectLocked(lock=True)
        self.viewbox.addItem(self.imageitem)

        self.plotitem = pg.PlotItem(viewBox=self.viewbox)
        self.plotitem.showGrid(x=True, y=True)
        self.plotitem.setAspectLocked(lock=False)
        self.win.addItem(self.plotitem)

        self.histgramlutitem = pg.HistogramLUTItem(self.imageitem)
        self.win.addItem(self.histgramlutitem)

        # TODO: QThread の適切な終了方法がわからん
        #       `QThread: Destroyed while thread is still running` と怒られてしまう
        # self.app.aboutToQuit.conncect(self.receiver.stop()) とかやると別の怒られが起きる
        # とりあえず放置。。。
        self.receiver = QThreadReceiver(addr, port)
        self.receiver.sigData.connect(self.draw)
        self.receiver.sigAttr.connect(self.set_attributes)
        self.receiver.sigError.connect(self.clear)
        self.receiver.start()

    def set_attributes(self, attrs: dict):
        if 'xlabel' in attrs:
            self.plotitem.setLabel('bottom', attrs['xlabel'])
        if 'ylabel' in attrs:
            self.plotitem.setLabel('left', attrs['ylabel'])
        if 'windowsize' in attrs:
            self.win.resize(*attrs['windowsize'])

    def clear(self):
        self.imageitem.clear()

    def draw(self, img):
        """
        args:
            - img, 2d array_like
        """
        vec = np.array(img)
        if len(vec.shape) != 2:
            self.clear()
            return

        self.imageitem.setImage(vec)



import sys
import argparse

from PySide2 import QtCore
from PySide2.QtWidgets import QApplication

from socket_plotter.api import DEFAULT_ADDR, DEFAULT_PORT_LINEPLOTTER
from socket_plotter.app.lineplotter import LinePlotter


def run(addr: str, port: int):
    _ = LinePlotter(addr, port)
    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QApplication.instance().exec_()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='launch a line plotter')
    parser.add_argument('--addr', default=DEFAULT_ADDR,
                        help=f'default={DEFAULT_ADDR}')
    parser.add_argument('--port', default=DEFAULT_PORT_LINEPLOTTER, type=int,
                        help=f'default={DEFAULT_PORT_LINEPLOTTER}')

    args = parser.parse_args()
    run(args.addr, args.port)
